// flirc - a smallweigth FLTK irc client
// Copyright (C) 2001 Georg Lukas <georg@boerde.de>
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// $Id: Fl_Irc_Output.cxx,v 1.3 2001/07/17 22:53:41 ge0rg Exp $

#include <FL/Fl.H>
#include <FL/Fl_Group.H>
#include <FL/fl_draw.H>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "Fl_Irc_Output.H"
#include "mystr.h"

/*class Fl_Irc_Output : public Fl_Widget {
	char **linebuf_;
	int linebufsize_;
	int linecount_;
	uchar textfont_;
	uchar textsize_;
};*/

Fl_Irc_Output::Fl_Irc_Output(int x,int y,int w,int h,const char *l = 0) : Fl_Widget(x,y,w,h,l) {
	linecount_ = 0;
	linebufsize_ = 100;
	linebuf_ = (char**)malloc(linebufsize_ * 4);
	linecol_ = (Fl_Color*)malloc(linebufsize_ * sizeof(Fl_Color));
	sb_ = NULL;
}

int Fl_Irc_Output::draw_line_rec(int x, int top_y, int w, int y, int d_, Fl_Color col, char *line) {
	char *rest_line;
	int sub_height, my_len;
	float lw;

	my_len = strlen(line);
	lw = fl_width(line);
	while ((int)(lw+0.5) > w) {
		my_len--;
		lw = fl_width(line, my_len);
	}
	if (my_len < strlen(line)) {
		// rec. call for drawing the rest
		sub_height = draw_line_rec(x, top_y, w, y, d_, col, line+my_len);
	} else
			sub_height = 0;
	//printf("x=%i y=%i l=%i s=%s\n", x, y-sub_height, my_len, line);
	if (y-sub_height > top_y) {
		// on screen check
		fl_color(FL_WHITE);
		fl_rectf(x, y-sub_height-textsize_, w, textsize_);
		fl_color(col);
		fl_draw(line, my_len, x, y-sub_height-d_);
	}
	return sub_height+textsize_;
}

void Fl_Irc_Output::draw() {
	int	x_ = x(), y_ = y(),
		w_ = w(), h_ = h(),
		draw_x = x_,
		draw_y = y_+h_,
		draw_d,
		line = linecount_ - 1;

	if (sb_ != NULL) line +=  - sb_size + sb_->value();

	/*if (damage() & FL_DAMAGE_ALL) {
		// draw background
		fl_color(FL_WHITE);
		fl_rectf(x_, y_, w_, h_);
		draw_box();
	}*/
	if (damage() & FL_DAMAGE_ALL) {
		// draw textlines
		fl_color(FL_BLACK);
		fl_font(textfont_, textsize_);
		fl_push_clip(x_, y_, x_+w_, y_+h_);
		draw_d = fl_descent();
		while ((draw_y-textsize_ >= y_) && (line >= 0)) {
			draw_y -= draw_line_rec(draw_x, y_, w_, draw_y, draw_d, linecol_[line], linebuf_[line]);
			line--;
		}
		if (draw_y-textsize_ > y_) {
			fl_color(FL_WHITE);
			fl_rectf(x_, y_, w_, draw_y-y_);
		}
		fl_pop_clip();
		draw_box();
	}
}

void Fl_Irc_Output::clear() {
	int i;
	for (i = 0; i<linecount_; i++) {
		free(linebuf_[i]);
	}
	linecount_ = 0;
	if (sb_ != NULL) sb_->value(0, 0, 0, 0);
	redraw();
}

void Fl_Irc_Output::add_line(char *line, Fl_Color col = FL_BLACK) {
	if (linecount_ == linebufsize_) {
		// free linebuf_[0], scroll up
		free(linebuf_[0]);
		for (int i = 1; i<linecount_; i++) {
			linebuf_[i-1] = linebuf_[i];
			linecol_[i-1] = linecol_[i];
		}
		linecount_--;
	}
	linebuf_[linecount_] = strnew(line);
	linecol_[linecount_++] = col;
	redraw();
	sb_size = linecount_ - (h() / textsize_);
	if (sb_size < 0) sb_size = 0;
	if (sb_ != NULL) sb_->value(sb_size, 20, 0, sb_size);
}
