// flirc - a smallweigth FLTK irc client
// Copyright (C) 2001 Georg Lukas <georg@boerde.de>
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// $Id: flirc.cxx,v 1.5 2001/07/17 22:53:41 ge0rg Exp $
// 
// TODO:
//	* implement build time modularity (suggested by Erik Ratcliffe)
//	* output formatting (mostly done, rework w/ dynaform[tm] dynamic formatter
//	* output color table
//	* auto connecting
//	* input classification
//	* string parser for input & output w/ support for full parameter list and
//    selected params (started for output)
//	* dynamic string handling (w/ deallocation)
//	* setup dialog
//	* ident daemon
//	* multiple channel/query-windows
//	* documentation (unimportant ;)

#include "globals.h"
#include "irc.h"
#include "mystr.h"
#include "setup_win.h"
#include "user.h"

#include <FL/Fl_Menu_Button.H>
#include "Fl_Irc_Output.H"


void irc_server();

Fl_Window *main_win;
Fl_Menu_Bar *main_menu;
Fl_Scrollbar *sb;
Fl_Irc_Output *output;
Fl_Input *cmdline;
Fl_Button *ok;
Fl_Menu_Button *history;

/********** input & output **********/
void clear() {
	output->clear();
}

void println(Fl_Color col, char *fmt, ...) {
	va_list ap;

	char buf[512];

	va_start(ap, fmt);
	vsnprintf(buf, 512, fmt, ap);
	va_end(ap);
	output->add_line(buf, col);
}

/********** callback functions **********/
void history_cb(Fl_Widget *w, void *data) {
	cmdline->insert(((Fl_Menu_*)w)->mvalue()->label());
}

void enter_cb(void) {
	if (cmdline->size() == 0) return;
	// [TODO] THIS IS A BAD BAD HACK ;)
	char *s = strnew(cmdline->value()),
	     *s_escaped = strnew(2*strlen(s));
	int i = 0, j = 0;
	while (s[i] != '\0') {
		if (s[i] == '/') {
			s_escaped[j] = '\\';
			j++;
		}
		s_escaped[j] = s[i];
		j++;
		i++;
	}
	s_escaped[j] = '\0';
	history->add(s_escaped, 0, history_cb, 0, 0);
	free(s_escaped);
	user_input(s);
	free(s);
	cmdline->value(NULL);
}

void scrollbar_cb(Fl_Widget *w, void *data) {
	output->redraw();
}

void disconnect_cb(Fl_Widget *w, void *data) {
	irc_disc(TITLE_FULL);
}

void connect_cb(Fl_Widget *w, void *data) {
	//println(COL_ERR, "connect_cb(%d; %s)", w, ((Fl_Menu_*)w)->mvalue()->label());
	irc_server(((Fl_Menu_*)w)->mvalue()->label());
}
void join_cb(Fl_Widget *w, void *data) {
	//println(COL_ERR, "join_cb(%d; %d)", w, data);
	irc_send("JOIN %s", ((Fl_Menu_*)w)->mvalue()->label());
}

void base_cb(void) {
	fl_alert("ALL YOUR BASE ARE BELONG TO US.");
}
void copy_cb(void) {
	cmdline->copy();
}
void cut_cb(void) {
	cmdline->copy();
	cmdline->cut();
}
void delete_cb(void) {
	cmdline->cut();
}
void paste_cb(void) {
	Fl::paste(*cmdline);
}

void flirc_quit(void) {
	//println(COL_STAT, NUM_STR "Quitting...");
	irc_disc(TITLE " going back to battery saving mode ;)");
	main_win->hide();
	setup_win->hide();
}


Fl_Menu_Item main_menu_items[] = {
	{ "&Server", 0, 0, 0, FL_SUBMENU },
		{ "irc.freenet.de",		0,		(Fl_Callback *)connect_cb,	0 },
		{ "irc.belwue.de",		0,		(Fl_Callback *)connect_cb,	0 },
		{ "irc.ham.de.euirc.net",0,		(Fl_Callback *)connect_cb,	0 },
		{ "irc.bre.de.euirc.net",0,		(Fl_Callback *)connect_cb,	0 },
		{ "irc.openprojects.net",0,	(Fl_Callback *)connect_cb,	0 },
		{ "tolkien.openprojects.net",0,	(Fl_Callback *)connect_cb,	0,
			 FL_MENU_DIVIDER },
		{ "&Setup...",		FL_ALT+'s',	(Fl_Callback *)setup_run_win,	0 },
		{ "&Disconnect",	FL_ALT+'d',	(Fl_Callback *)disconnect_cb,	0,
			 FL_MENU_DIVIDER },
		{ "&Quit",			FL_ALT+'q',	(Fl_Callback *)flirc_quit },
	{ 0 },
	{ "&Channel", 0, 0, 0, FL_SUBMENU },
		{ "#agenda",		0,			(Fl_Callback *)join_cb,	0 },
		{ "#flirc",			0,			(Fl_Callback *)join_cb,	0 },
		{ "#icfs",			0,			(Fl_Callback *)join_cb,	0,
			FL_MENU_DIVIDER },
		{ "#germany",		0,			(Fl_Callback *)join_cb,	0 },
		{ "#linux",			0,			(Fl_Callback *)join_cb,	0 },
		{ "#magdeburg",		0,			(Fl_Callback *)join_cb,	0 },
	{ 0 },
	{ "&Edit", 0, 0, 0, FL_SUBMENU },
		//{ "&Undo",			FL_ALT+'z',	(Fl_Callback *)undo_cb, 0,
		//	FL_MENU_DIVIDER },
		{ "Cu&t",			FL_ALT+'x',	(Fl_Callback *)cut_cb },
		{ "&Copy",			FL_ALT+'c',	(Fl_Callback *)copy_cb },
		{ "&Paste",			FL_ALT+'v',	(Fl_Callback *)paste_cb },
		{ "&Delete",		0,			(Fl_Callback *)delete_cb },
	{ 0 },
	{ "    ", 0, (Fl_Callback *)base_cb },
	{ 0 }
};

Fl_Menu_Item aliases[] = {
	{ "/NICK ",  0, history_cb },
	{ "/JOIN ", 0, history_cb },
	{ "/PART ", 0, history_cb },
	{ "/DISC ", 0, history_cb },
	{ "/QUIT ", 0, history_cb, NULL, FL_MENU_DIVIDER },
	{ 0 }
};

int main(int argc, char **argv) {
	// init basic data
	//printf("Setting default values...\n");
	irc_nick = strnew(getenv("LOGNAME"));
	irc_user = strnew(getenv("LOGNAME"));
	//irc_nick = strnew(cuserid(0));
	//irc_user = strnew(cuserid(0));
	irc_name = strnew(TITLE_FULL);
	irc_target = NULL;

	// Main Window
	//printf("Creating main window...\n");
	main_win = new Fl_Window(160,240, TITLE);

	main_menu = new Fl_Menu_Bar(0, 0, 160, 15);
	main_menu->box(FL_NO_BOX);
	main_menu->menu(main_menu_items);
	main_menu->textfont(MENU_FONT);
	main_menu->textsize(MENU_FONTSIZE);

	output = new Fl_Irc_Output(0,15,152,211);
	//output->box(FL_DOWN_FRAME);
	output->textfont(IRC_FONT);
	output->textsize(IRC_FONTSIZE);
	sb = new Fl_Scrollbar(152, 15, 8, 211);
	sb->callback(scrollbar_cb);
	output->scrollbar(sb);


	/*connect = new Fl_Menu_Button(0, 0, 40, 13, "&Connect");
	connect->labelfont(IRC_FONT);
	connect->labelsize(IRC_FONTSIZE);
	connect->menu(servers);*/

	cmdline = new Fl_Input(14,227,146,14,"");
	cmdline->textfont(IRC_FONT);
	cmdline->textsize(IRC_FONTSIZE);
	cmdline->callback((Fl_Callback *)enter_cb);
	cmdline->when(FL_WHEN_ENTER_KEY);

	history = new Fl_Menu_Button(0,227,14,13,"@8->");
	history->menu(aliases);
	history->labeltype(FL_SYMBOL_LABEL);
	history->labelsize(8);
	cmdline->textfont(IRC_FONT);
	history->textsize(8);
	//history->callback((Fl_Callback *)enter_cb);

	//ok = new Fl_Button(146,227,14,13,"@1->");
	//ok->labeltype(FL_SYMBOL_LABEL);
	//ok->labelsize(8);
	//ok->callback((Fl_Callback *)enter_cb);

	println(FL_BLACK, TITLE " Copyright (C) 2001 G. Lukas");
	println(FL_BLACK, NUM_STR "flirc comes with ABSOLUTELY NO WARRANTY; It is free software, and can be copied or modified under the terms of the GNU GPL (for details read the LICENSE file).");
	main_win->end();
	main_win->resizable(output);
	main_win->show(argc, argv);

	setup_create_win();
	//setup_run_win();

	//irc_server("irc.freenet.de");

	return Fl::run();
}

