// flirc - a smallweigth FLTK irc client
// Copyright (C) 2001 Georg Lukas <georg@boerde.de>
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// $Id: irc.cxx,v 1.5 2001/07/17 22:53:41 ge0rg Exp $

#include "globals.h"
#include "builddate.h"
#include "mystr.h"
// Alles was mit dem Socket zu tun hat, kommt hier rein....
char *irc_servername, *irc_nick, *irc_user, *irc_name, *irc_target;
int line_num = 0, irc_serverport;

enum Connect_State {
	CS_DISCONNECTED,
	CS_RESOLVING, // todo
	CS_CONNECTING,
	CS_CONNECTED,
	CS_AUTHSENT,
	CS_AUTHOK,
	CS_RUNNING,
	CS_QUITTING };

int irc_sock, irc_state;

void irc_send(const char* fmt, ...);
void irc_connected();

typedef char* String;

struct Irc_Message {
	char *src;
	char *src_nick;
	char *src_uhost;
	char *cmd_str;
	int cmd;
	bool ctcp;
	int pcnt;
	char **p;
	char *tail; // alle ausser p[0]; bei CTCP: alle ausser p[1]
};

void free_im(Irc_Message *im) {
	int i;
	free(im->src); im->src = NULL;
	free(im->src_nick); im->src_nick = NULL;
	free(im->src_uhost); im->src_uhost = NULL;
	free(im->cmd_str); im->cmd_str = NULL;
	if (im->pcnt>0) for (i=0; i<im->pcnt; i++) {
		free(im->p[i]); im->p[i] = NULL;
	}
	free(im->p); im->p = NULL;
	im->pcnt = 0;
	free(im->tail); im->tail = NULL;
}


// rfc1459.txt line 432 
// [:(server|nick[!user@host]) ]command{ param}[ :trailing]
void parse_im(Irc_Message *im, char *line) {
	char *ptr1, *ptr2, *ptr3;

	im->ctcp = false;
	ptr1 = line;
	if (*ptr1 == ':') {
		// parse prefix
		ptr1++;
		// save full prefix to src
		im->src = irctoken(&ptr1, &im->ctcp);

		// save nick part to src_nick
		ptr3 = strchr(im->src, '!');
		if (ptr3 != NULL) {
			im->src_nick = strnew(im->src, (ptr3-im->src));
			// save user@host part to src_uhost
			im->src_uhost = strnew(ptr3+1);
		} else {
			// no ! found - copy im->src to nick
			im->src_nick = strnew(im->src);
		}
	}

	// get command
	im->cmd_str = irctoken(&ptr1, &im->ctcp);
	im->cmd = atol(im->cmd_str);

	im->pcnt = 0;
	im->p = NULL;
	while (ptr1 != NULL) {
		// get another argument
		// only set it if not CTCP
		if ((im->pcnt == 1) && (ptr1[0] != '\000') && (ptr1[1] != '\001')) 
			im->tail = strnew(ptr1);
		else
		// get CTCP argument
		if (im->ctcp && (im->pcnt == 2)) {
			im->tail = strnew(ptr1);
		}

		im->p = (char**)realloc(im->p, sizeof(String)*(im->pcnt+1));
		im->p[im->pcnt] = irctoken(&ptr1, &im->ctcp);
		im->pcnt++;
	}

	/*printf("%s %s %s %s %i %i\n", im->src, im->src_nick, im->src_uhost,
		im->cmd_str, im->cmd, im->pcnt);/**/
}


/*
struct Serv_Num_Msg {
	int num;
	int refnum; // if !=0 then get fmtstr from self[refnum]
	char *fmtstr;
	(void *)(void) callback;
} num_msgs[1000] = {
	{   1,   0, NUM_STR "$1" },
	{   2,   1 },
	{   3,   1 },
	{   4,   1 },
	{ 251,   1 },
	{ 255,   1 },
	{ 256,   1 },
	{ 256,   1 },*/

void interpret_ctcp(Irc_Message im) {
	if (strcmp(im.p[1], "PING") == 0) {
		println(COL_STAT, NUM_STR "[%s PING] %s (%s)", im.src_nick, im.tail, im.p[0]);
		irc_send("NOTICE %s :\001PING %s\001", im.src_nick, im.tail);
	} else
	if (strcmp(im.p[1], "VERSION") == 0) {
		println(COL_STAT, NUM_STR "[%s VERSION] (%s)", im.src_nick, im.p[0]);
		irc_send("NOTICE %s :\001VERSION " TITLE_FULL " (build " BUILD_DATE ")\001", im.src_nick);
	} else
	if (strcmp(im.p[1], "ACTION") == 0) {
		println(COL_STAT, "* %s:%s %s", im.src_nick, im.p[0], im.tail);
	}
}

void interpret_server(char *line) {
	Irc_Message im = { 0, 0, 0, 0, 0, 0, 0, 0 };

	//printf("%s\n", line);
	// excremental code
	parse_im(&im, line);

	// fallback check for evil "NICK change during connect" bug
	if (im.cmd != 433) irc_state = CS_AUTHOK;

	switch (im.cmd) {

	// SERVER GREETINGS
	case 1: case 2: case 3: case 4: // greetings w/o formatting
	case 251: case 255: case 265: case 266: //server status
		println(FL_BLACK, NUM_STR "%s", im.p[1]); break;
	case 5: break; // TOKEN WATCH=...

	case 252: case 253: case 254: //server status w/ integer
		println(FL_BLACK, NUM_STR "%s %s", im.p[1], im.p[2]); break;

	// WHOIS
	case 311: // whois user
		println(COL_STAT, NUM_STR "%s (%s@%s) is %s", im.p[1], im.p[2], im.p[3], im.p[im.pcnt-1]); break;
	case 312: // whois server
		println(COL_STAT, NUM_STR "%s on %s (%s)", im.p[1], im.p[2], im.p[im.pcnt-1]); break;
	case 313: // whois ircop
		println(COL_STAT, NUM_STR "%s %s", im.p[1], im.p[im.pcnt-1]); break;
	case 314: // whowas(?) user
		println(COL_STAT, NUM_STR "%s (%s@%s) %s", im.p[1], im.p[2], im.p[3], im.p[im.pcnt-1]); break;
	case 317: // whois idle
		println(COL_STAT, NUM_STR "%s is %i %s", im.p[1], im.p[2], im.p[im.pcnt-1]); break;
	// WHO
	case 352:
		println(COL_STAT, NUM_STR "%s (%s@%s) is %s on %s using %s [%s]", im.p[5], im.p[2], im.p[3], im.p[7], im.p[1], im.p[4], im.p[6]);
		break;
	case 318: // end of whois
	case 369: // end of whowas
	case 315: // end of who
		break;
	case 319: // whois idle
		println(COL_STAT, NUM_STR "%s: %s", im.p[1], im.p[im.pcnt-1]); break;


	case 331: // no topic
		println(COL_STAT, NUM_STR "%s has no topic.", im.p[1]); break;
	case 332: // topic
		println(COL_STAT, NUM_STR "%s topic: '%s'", im.p[1], im.p[2]); break;
	case 333: // topic by
		println(COL_STAT, NUM_STR "%s topic set by %s at %s", im.p[1], im.p[2], im.p[3]); break;
	case 353: // nick list
		println(COL_STAT, NUM_STR "%s users: %s", im.p[2], im.p[im.pcnt-1]); break;
	case 366: // End Of nick list
		break;

	case 372: case 375: case 376: // MOTD - ignore
		break;

	case 401: case 402: case 403: case 404: case 405: case 406: case 407:
// unknown user, server, channel, cantsend, 2many#s, wasnonick, 2manyTargets
	case 409: case 411: case 412:
		println(COL_ERR, NUM_STR "%s", im.p[1]); break;
		break;
	case 413: case 414:
	case 421: // unknown command
	case 422: // missing motd
	case 423: // no admin
	case 424: // file error
	case 431: // no nick given
	case 432: // wrong nick given
	case 442: // not on channel
	case 482: // not a chanop
		println(COL_ERR, NUM_STR "%s: %s", im.p[1], im.p[2]); break;
	case 441: // user not in channel
	case 443: // user on channel
		println(COL_ERR, NUM_STR "%s: %s %s", im.p[1], im.p[2], im.p[3]); break;
	case 471: // +l
		println(COL_ERR, NUM_STR "%s: Channel full!", im.p[1]); break;
	case 473: // +i
		println(COL_ERR, NUM_STR "%s: Invite only!", im.p[1]); break;
	case 474: // +b
		println(COL_ERR, NUM_STR "%s: You are banned!", im.p[1]); break;
	case 475: // +k
		println(COL_ERR, NUM_STR "%s: Bad key!", im.p[1]); break;
	case 433: // nick given already in use
		// [WICHTIG: hier Nick-Abfrage-Dialog platzieren]
		println(COL_ERR, NUM_STR "%s: %s", im.p[1], im.p[2]); break;
		break;

	default:
		if (strcmp(im.cmd_str, "PING") == 0) {
			irc_send("PONG %s", im.p[0]);
			println(COL_STAT, NUM_STR "ping? pong! (%s)", im.p[0]);
		} else
		if (strcmp(im.cmd_str, "ERROR") == 0) {
			println(COL_ERR, NUM_STR "%s", im.p[0]);
			irc_disc(NULL);
		} else
		if (strcmp(im.cmd_str, "PRIVMSG") == 0) {
			if (im.ctcp) interpret_ctcp(im);
			else {
				println(COL_MSG(line_num), "<%s:%s> %s", im.src_nick, im.p[0], im.p[1]);
				line_num = 1-line_num;
			}
		} else
		if (strcmp(im.cmd_str, "NOTICE") == 0) {
			println(FL_BLACK, "-%s:%s- %s", im.src_nick, im.p[0], im.p[1]);
		} else
		if (strcmp(im.cmd_str, "INVITE") == 0) {
			println(COL_STAT, NUM_STR "%s invites you to %s",
				im.src_nick, im.p[1]);
		} else
		if (strcmp(im.cmd_str, "JOIN") == 0) {
			if (strcmp(irc_nick, im.src_nick) == 0)
				irc_target = strrealloc(irc_target, im.p[0]);
			println(COL_STAT, NUM_STR "%s (%s) has joined %s",
				im.src_nick, im.src_uhost, im.p[0]);
		} else
		if (strcmp(im.cmd_str, "PART") == 0) {
			if (strcmp(irc_nick, im.src_nick) == 0) {
				free(irc_target); irc_target = NULL;
			}
			println(COL_STAT, NUM_STR "%s (%s) has left %s (%s)",
				im.src_nick, im.src_uhost, im.p[0], im.p[1]);
		} else
		if (strcmp(im.cmd_str, "KICK") == 0) {
			if (strcmp(irc_nick, im.p[1]) == 0) {
				free(irc_target); irc_target = NULL;
			}
			println(COL_STAT, NUM_STR "%s kicked %s from %s (%s)",
				im.src_nick, im.p[1], im.p[0], im.p[2]);
		} else
		if (strcmp(im.cmd_str, "QUIT") == 0) {
			println(COL_STAT, NUM_STR "%s (%s) has quit IRC (%s)",
				im.src_nick, im.src_uhost, im.p[0]);
		} else
		if (strcmp(im.cmd_str, "MODE") == 0) {
			println(COL_STAT, NUM_STR "%s changed mode on %s to: %s",
				im.src_nick, im.p[0], im.tail);
		} else
		if (strcmp(im.cmd_str, "TOPIC") == 0) {
			println(COL_STAT, NUM_STR "%s changed topic on %s to: %s",
				im.src_nick, im.p[0], im.tail);
		} else
		if (strcmp(im.cmd_str, "NICK") == 0) {
			if (strcmp(irc_nick, im.src_nick) == 0) // my nick changed?
				irc_nick = strrealloc(irc_nick, im.p[0]);
			println(COL_STAT, NUM_STR "%s is now known as %s", im.src_nick, im.p[0]);
		} else 
		{
			//println(FL_BLACK, "(%s)%s %s %s", im.src, im.cmd_str, im.p[0], im.tail);
			println(COL_ERR, "%s", line);
			//printf("%s\n", line);
		}

	}

	free_im(&im);
}


char socket_read_buf[2049];
int socket_read_size;

void socket_read(int s, void *param) {
	char *act, *next;
	int d_size;
	if (irc_state < CS_CONNECTED ) {
		println(COL_ERR, NUM_STR "socket_read(): Not connected!");
		return;
	}
	d_size = read(irc_sock, &socket_read_buf[socket_read_size], 2048-socket_read_size);
	if (d_size == -1) {
		if (errno == EAGAIN) return;
		println(COL_ERR, "read(irc_sock): %s", strerror(errno));
		irc_disc(NULL);
		return;
	}
	socket_read_size += d_size;
	socket_read_buf[socket_read_size]='\0';
	act = (char*)&socket_read_buf;
	do {
		next = strchr(act, '\r');
		if (next == NULL) {
			// copy data to buffer start for reuse
			socket_read_size = strlen(act);
			memcpy(&socket_read_buf, act, socket_read_size);
		} else {
			// parse rest of string
			*next = 0;
			next += 2;
			interpret_server(act);
			act = next;
		}
	} while (next != NULL);
}

/*void socket_write(int s, void *param) {
	if (irc_state >= CS_CONNECTED) return;
	if (irc_state == CS_CONNECTING) {
		irc_state = CS_CONNECTED;
		irc_connected();
	}
}*/

void socket_except(int s, void *param) {
	printf(NUM_STR "socket except called! %i\n", s);
	println(COL_ERR, NUM_STR "Exception caught. Disconnecting!");
	irc_disc(NULL);
}

void irc_send(const char *fmt, ...) {
	va_list ap;

	char sendbuf[515]; // 512 + \cr\lf + \0
	int res, res2;

	if (irc_state < CS_CONNECTED)
		return;

	//vprintf(fmt, ap);
	va_start(ap, fmt);
	vsnprintf(sendbuf, 512, fmt, ap);
	va_end(ap);
	strncat(sendbuf, "\r\n", 2);
	res = write(irc_sock, sendbuf, strlen(sendbuf));
	if (res == -1) {
		println(COL_ERR, NUM_STR "send(): %s", strerror(errno));
		irc_disc(NULL);
	} else {
		fsync(irc_sock);
		//printf("%s\n", line);
	}
}

void irc_connected() {
	if (irc_state != CS_CONNECTED) {
		println(COL_ERR, NUM_STR "connected() called w/o reason.");
		return;
	}
	irc_send("NICK %s", irc_nick);
	irc_send("USER %s 0 0 :%s", irc_user, irc_name);
	irc_state = CS_AUTHSENT;
}

void irc_disc(const char *quitmsg) {
	if (irc_state == CS_DISCONNECTED) {
		println(COL_ERR, NUM_STR "Not connected.");
		return;
	}
	if ((irc_state == CS_QUITTING) || (irc_state < CS_CONNECTED) || (quitmsg == NULL)) {
		close(irc_sock);
		Fl::remove_fd(irc_sock);
		free(irc_target); irc_target = NULL;
		irc_state = CS_DISCONNECTED;
		println(COL_ERR, NUM_STR "Disconnected");
	} else {
		irc_send("QUIT :%s", quitmsg);
	}
}

bool irc_setnick(const char *newnick) {
	if (irc_state >= CS_AUTHSENT) {
		irc_send("NICK %s", newnick);
		// when not fully connected, change local nick version too
		return (irc_state >= CS_AUTHOK);
	} else return false;
}


void irc_server(const char *server, int port) {
	sockaddr_in serv;
	struct hostent *serv_ns;
	int flags;
	if (irc_state != CS_DISCONNECTED) {
		println(COL_ERR, NUM_STR "Already connected.");
		return;
	}
	if ((irc_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
		println(COL_ERR, NUM_STR "socket(): %s", strerror(errno));
		return;
	}
	if ((serv_ns = gethostbyname(server)) !=NULL) {
		memcpy((char*)&serv.sin_addr, serv_ns->h_addr, serv_ns->h_length);
	} else {
		println(COL_ERR, NUM_STR "gethostbyname(%s): %s", server, strerror(errno));
		return;
	}
	serv.sin_port=htons(port);
	serv.sin_family = AF_INET;

	socket_read_size = 0;
	Fl::add_fd(irc_sock, FL_EXCEPT, socket_except, NULL);
	flags = fcntl(irc_sock, F_GETFL, 0);
	/*if ( (flags == -1) || (fcntl(irc_sock, F_SETFL, flags  & ~O_NONBLOCK) == -1)) {
		println(COL_ERR, NUM_STR "fcntl(): %s", strerror(errno));
		return;
	}*/
	if (connect(irc_sock,(struct sockaddr *)&serv,16)==-1) {
		if (errno == EINPROGRESS) {
			println(COL_STAT, NUM_STR "Connecting...");
			irc_state = CS_CONNECTING;
			return;
		}
		println(COL_ERR, NUM_STR "connect(): %s", strerror(errno));
		return;
	}
	Fl::add_fd(irc_sock, FL_READ, socket_read, NULL);
	//Fl::add_fd(irc_sock, FL_WRITE, socket_write, NULL);
	println(COL_STAT, NUM_STR "Connected to server");
	irc_state = CS_CONNECTED;
	irc_target = NULL;
	irc_connected();
}

void irc_server(const char *server) {
	char *s, *p;
	int port;

	p = strchr(server, ':');
	if (p == NULL) {
		// no : found, ergo using default port
		irc_server(server, IRC_PORT);
	} else {
		s = strnew(server, (p-server));
		p++;
		port = atol(p);
		irc_server(s, port);
		free(s);
	}
}
