// flirc - a smallweigth FLTK irc client
// Copyright (C) 2001 Georg Lukas <georg@boerde.de>
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// $Id: user.cxx,v 1.3 2001/07/17 22:53:41 ge0rg Exp $

#include "setup_win.h"
#include "irc.h"
#include "globals.h"
#include "mystr.h"

// predefinition of used functions
void user_ctcp();
void user_disc();
void user_help();
void user_me();
void user_msg();
void user_notice();
void user_setnick();
void user_server();
void user_quit();

void user_join();
void user_part();

typedef void (User_Cmd)(void);
struct User_Cmd_List {
	const char *name;
	User_Cmd *cmd;
	int min_params;
	const char *info;
};

User_Cmd_List user_cmds[] = {
	{ "CLEAR", clear, 0, "clears the screen buffer" },
	{ "CTCP", user_ctcp, 2, "sends a CTCP request to s.o." },
	{ "DISC", user_disc, 0, "disconnects from server" },
	{ "JOIN", user_join, 1, "joins a channel" },
	{ "HELP", user_help, 0, "shows this message" },
	{ "ME", user_me, 1, "send a CTCP ACTION to the channel" },
	{ "MSG", user_msg, 2, "sends a message to s.o." },
	{ "NICK", user_setnick, 1, "changes your nick" },
	{ "NOTICE", user_notice, 2, "sends a notice to s.o." },
	{ "PART", user_part, 0, "parts a channel" },
	{ "QUIT", user_quit, 0, "quits the client" },
	{ "SERVER", user_server, 1, "connects to a given server" },
	{ "SETUP", setup_run_win, 0, "opens the setup dialog" },
	{ NULL }
};

char *cmd, *p1, *p2, *pX, *pall;

void user_help() {
	int i = 0;
	while (user_cmds[i].name != NULL) {
		if ((p1==NULL) || (strcmp(p1, user_cmds[i].name) == 0)) {
			println(FL_BLACK, NUM_STR "%s: %s", user_cmds[i].name, user_cmds[i].info);
		}
		i++;
	}
}

void user_setnick() {
	if (!irc_setnick(p1)) {
		irc_nick = strrealloc(irc_nick, p1);
		println(COL_STAT, NUM_STR "Your new nick: %s", irc_nick);
	}
}

void user_server() {
	// accepts "server port" or "server:port"
	if (p2!=NULL) {
		int port;
		port = atol(p2);
		irc_server(p1, port);
	} else irc_server(p1);
}

void user_disc() {
	irc_disc(pall);
}

void user_quit() {
	irc_disc(pall);
	flirc_quit();
}

void user_me() {
	if (irc_target != NULL) {
		irc_send("PRIVMSG %s :\1ACTION %s\1", irc_target, pall);
		println(COL_STAT, "* %s %s", irc_nick, pall);
	} else
		println(COL_ERR, NUM_STR "Please JOIN a channel first");
}
void user_msg() {
	irc_send("PRIVMSG %s :%s %s", p1, p2, pX);
		println(COL_STAT, "-> *%s* %s %s", p1, p2, pX);
}
void user_ctcp() {
	irc_send("PRIVMSG %s :\001%s %s\001", p1, p2, pX);
		println(COL_STAT, "-> *%s* %s %s", p1, p2, pX);
}
void user_notice() {
	irc_send("NOTICE %s :%s %s", p1, p2, pX);
		println(COL_STAT, "-> *%s* %s %s", p1, p2, pX);
}
void user_join() {
	irc_send("JOIN %s", p1);
}
void user_part() {
	if (p2 == NULL) p2 = TITLE;
	if (pX)
		irc_send("PART %s :%s %s", p1, p2, pX);
	else
	if (p2)
		irc_send("PART %s :%s", p1, p2);
	else
	if (p1)
		irc_send("PART %s", p1);
	else
		irc_send("PART %s", irc_target);
}

/********** command handling **********/
void user_interpret(const char *input) {
	char buf[128], buf2[128];
	int i;

	// nicht crashen wenn nur / kommt
	if (input[0] == '\0') return;
	// alles was mit // startet, an den Server quoten
	if (input[0] == CMD_CHAR) {
		irc_send("%s", input+1);
		return;
	}

	strcpy(buf, input);
	cmd = strtok(buf, " ");
	if (strlen(cmd)<strlen(input)) {
		pall = strnew(cmd + strlen(cmd)+1);
	} else pall = NULL;
	p1 = strtok(NULL, " ");
	p2 = strtok(NULL, " ");
	pX = strtok(NULL, "\0");

	i=0;
	while (cmd[i] != '\0')
		cmd[i++] = toupper(cmd[i]);

	i=0;
	while (user_cmds[i].name != NULL) {
		if (strcmp(cmd, user_cmds[i].name) == 0) {
			// command found, [check length] 
			if (user_cmds[i].cmd != NULL) {
			user_cmds[i].cmd();
			} else println(COL_ERR, NUM_STR "Not yet implemented: %s", cmd);
			free(pall);
			return;
		}
		i++;
	}
	// command not locally supported - sending to server
	irc_send("%s", input);
	//println(COL_ERR, NUM_STR "Unknown cmd: %s", cmd);
	free(pall);
}

void user_input(const char *input) {
	if (input[0] == CMD_CHAR) {
		user_interpret(input+1);
	} else {
		if (irc_target != NULL) {
			// send it to the act. channel
			irc_send("PRIVMSG %s :%s", irc_target, input);
			println(FL_BLUE, "> %s", input);
		} else
			// not on a channel
			println(COL_ERR, NUM_STR "Not on a channel!");
	}
}

